-- Scorchio2!
-- A multi-mob vulnerability manager
-- ennui@bloodhoof-eu
-- Email: ennuilg@gmail.com

Scorchio2 = LibStub("AceAddon-3.0"):NewAddon("Scorchio2", "AceEvent-3.0", "AceConsole-3.0", "AceTimer-3.0", "LibBars-1.0", "LibSink-2.0")
local L = LibStub:GetLibrary("AceLocale-3.0"):GetLocale("Scorchio2")
local SM = LibStub("LibSharedMedia-3.0")
local AceConfig = LibStub("AceConfig-3.0")
local AceConfigDialog = LibStub("AceConfigDialog-3.0")

-- Constants for mobTable
local MOBNAME = 1
local REFRESHTIME = 2
local STACK = 3
local CASTS = 4
local TIMER = 5
local BAR = 6

local T = { "a", "b", "c" }
 
-- Spell Constants
local SPELLIDS = {
        { [2948]=true, [8444]=true, [8445]=true, [8446]=true, [10205]=true, [10206]=true, [10207]=true, [27903]=true, [27074]=true, [42858]=true, [42859]=true   }, -- Scorch
        { [116]=true, [205]=true, [837]=true, [7322]=true, [8406]=true, [8407]=true, [8408]=true, [10179]=true, [10180]=true, [10181]=true, [25304]=true, [27071]=true, [27072]=true, [38697]=true, [42841]=true, [42842]=true, -- Frostbolt
          [122]=true, [865]=true, [6131]=true, [10230]=true, [27088]=true, [42917]=true, -- Frost Nova
          [120]=true, [8492]=true, [10159]=true, [10160]=true, [10161]=true, [27087]=true, [42930]=true, [42931]=true, -- Cone of Cold
          [30455]=true, [42913]=true, [42914]=true, -- Ice Lance
          [44614]=true, [47610]=true -- Frostfire Bolt
        },
        { --[[ [44461]=true, [55361]=true, [55362]=true --]] }, -- Living Bomb
        -- { [31589]=true } -- Slow
    }
local VULNERABILITIES = {
        { [22959]=true, -- Improved Scorch
          [38715]=true -- Fire Vulnerability
        },
        { [12579]=true -- Winter's Chill
        },
        { [44457]=true, -- Living Bomb I
          [55359]=true, -- Living Bomb II
          [55360]=true -- Living Bomb III
        }
    }
local VULNDATA = {
    { -- Scorch
        message = L["Fire Vulnerability Expired on %s"],
        warning = L["Rescorch %s!"],
        duration = 30,
        maxstacks = 5,
        unique = false,
        valid = 4294967295,
        icon = "Interface\\Icons\\Spell_Fire_SoulBurn" 
    },
    { -- Winter's Chill
        message = L["Winter's Chill Expired on %s"],
        warning = L["Reapply Winter's Chill on %s!"],
        duration = 15,
        maxstacks = 5,
        unique = false,
        valid = 4294967295,
        icon = "Interface\\Icons\\Spell_Frost_ChillingBlast"
    },
    { -- Living Bomb
        message = L["Living Bomb Exploded on %s"],
        warning = L["%s Is About To Blow!"],
        duration = 12,
        maxstacks = 1,
        unique = true,
        valid = COMBATLOG_OBJECT_AFFILIATION_MINE,
        icon = "Interface\\Icons\\Ability_Mage_LivingBomb"
    }
}
    
-- Utility Functions
local function Debug(class, msg)
	local text = "DEBUG: " .. msg
	if Scorchio2.db.profile.debug[class] then
		Scorchio2:Print(text)
	end
end
Scorchio2.Debug = Debug
    
local function Colours(pack, alpha)
    local r, g, b = unpack(pack)
    return r, g, b, alpha
end
Scorchio2.Colours = Colours
    
local function GetSMIndex(t, value)
	for k, v in pairs(SM:List(t)) do
	if v == value then
		return k
		end
	end
	return nil
end


-- Defaults and Options Tables
local defaults = {
	profile = {
		baroptions = {
			position = { 
				main = { }
			},
			growth = true,
			scale = 1.0,
			fadetime = -1,
			flash = false,
			flashtime = 6,
			iconshow = true,
			iconside = "LEFT",
			raidiconshow = false,
			raidiconpos = -10,
            texture = "Blizzard",
			width = 200,
			height = 18,
            maxbars = 0,
			font = "Arial Narrow",
			fontsize = 12,
            a = {
                track = true,
                show = true,
                warningtime = 6,
                soundson = true,
                warningsound = "Bell",
                expiredsound = "Info",
                bar = "$s x $m",
                fg = { 0, 1, 0 },
                bg = { 0.2, 0.2, 0.2 },
                },
            b = {
                track = true,
                show = true,
                warningtime = 5,
                soundson = true,
                bar = "$s x $m",
                warningsound = "Bell",
                expiredsound = "Info",
                fg = { 0, 0, 1 },
                bg = { 0.2, 0.2, 0.2 },
                },
            c = {
                track = true,
                show = true,
                warningtime = 0,
                soundson = true,
                bar = "$m",
                warningsound = "Bell",
                expiredsound = "Info",                
                fg = { 1, 0, 0 },
                bg = { 0.2, 0.2, 0.2 },
                },
		},
		sinkoptions = { },
        debug = {
            duplicates = false,
            sync = false,
            mob = false,
            code = false,
            timers = false,
            log = false
        },
	},
}


local options = {
	type = 'group',
	args = {
        baroptions = {
			type = 'group',
			name = L["Bar Options"],
			desc = L["Bar Options"],
			args = {
                toggle = {
                    type = 'execute',
                    name = L["Toggle Anchors"],
                    desc = L["Shows/Hides the draggable anchors"],
                    func = function()
                            Scorchio2:ToggleAnchors()
                    end,
                },
                test = {
                    type = 'execute',
                    name = L["Test"],
                    desc = L["Shows test bars"],
                    func = function()
                               Scorchio2:RunTest()
                           end,
                },
				barwidth = {
					type = 'range',
					name = L["Width"],
					desc = L["Change the width of the bars"],
					order = 120,
					get = function()
						return Scorchio2.db.profile.baroptions.width
					end,
					set = function(info, v)
						Scorchio2.db.profile.baroptions.width = v
						Scorchio2:UpdateAnchors()                   
					end,
					min = 24,
					max = 512,
					step = 8,
				},
				barheight = {
					type = 'range',
					name = L["Height"],
					desc = L["Change the height of the bars"],
					order = 121,
					get = function()
						return Scorchio2.db.profile.baroptions.height
					end,
					set = function(info, v)
						Scorchio2.db.profile.baroptions.height = v
						Scorchio2:UpdateAnchors()
					end,
					min = 6,
					max = 64,
					step = 2,
				},
				scale = {
					type = 'range',
					name = L["Scale"],
					desc = L["Change the scale of the bars"],
					order = 122,
					min = 0.5,
					max = 2,
					get = function()
						return Scorchio2.db.profile.baroptions.scale
					end,
					set = function(info, v)
						Scorchio2.db.profile.baroptions.scale = v
						Scorchio2:UpdateAnchors()
					end
				},
				growth = {
					type = 'toggle',
					name = L["Grow Up"],
					desc = L["If checked, bar will stack above the anchor, not below"],
					order=123,
					get = function()
						return Scorchio2.db.profile.baroptions.growth
					end,
					set = function(info, v)
						Scorchio2.db.profile.baroptions.growth = v
						Scorchio2:UpdateAnchors()
					end
				},
                maxbars = {
                    type = 'range',
                    name = L["Max Bars"],
                    desc = L["Maximum number of bars to show at any one time. Set to 0 to show all bars."],
                    order = 124,
                    min = 0,
                    max = 10,
                    step = 1,
                    get = function()
                        return Scorchio2.db.profile.baroptions.maxbars
                    end,
                    set = function(info, v)
                        Scorchio2.db.profile.baroptions.maxbars = v
                        Scorchio2:UpdateAnchors()
                    end
                },
				fontsize = {
					type = 'range',
					name = L["Font Size"],
					desc = L["Change the font size"],
					order = 125,
					min = 7,
					max = 24,
					step = 1,
					get = function()
					return Scorchio2.db.profile.baroptions.fontsize
					end,
					set = function(info, v)
					Scorchio2.db.profile.baroptions.fontsize = v
					Scorchio2:UpdateAnchors()
					end
                		},
                font = {
					type = 'select',
                    name = L["Font"],
                    desc = L["Set the bar font"],
                    order = 126,
					values = SM:List("font"),
					get = function(info)
						return GetSMIndex("font", Scorchio2.db.profile.baroptions.font)
					end,
					set = function(info, v)
						Scorchio2.db.profile.baroptions.font = SM:List("font")[v]
						Scorchio2:UpdateAnchors()
					end,
				},
				texture = {
                    type = 'select',
					name = L["Texture"],
					desc = L["Change the texture of the bars"],
					order = 127,
					values = SM:List("statusbar"),
					get = function() 
						return GetSMIndex("statusbar", Scorchio2.db.profile.baroptions.texture) 
					end,
					set = function(info, v)
						Scorchio2.db.profile.baroptions.texture = SM:List("statusbar")[v]
						Scorchio2:UpdateAnchors()
					end,
				},
                a = {
                    type = 'group',
                    name = L["Scorch Bar Options"],
                    desc = L["Scorch Bar Options"],
                    args = {
                        track = {
                            type = 'toggle',
                            name = L["Track Scorch"],
                            desc = L["Track Scorch"],
                            order = 131,
                            get = function()
                                return Scorchio2.db.profile.baroptions.a.track
                            end,
                            set = function(info, v)
                                Scorchio2.db.profile.baroptions.a.track = v
                                Scorchio2:UpdateAnchors()
                            end
                        },                        
                        show = {
                            type = 'toggle',
                            name = L["Show Scorch Bars"],
                            desc = L["Show Scorch Bars"],
                            order = 132,
                            disabled = function()
                                return not Scorchio2.db.profile.baroptions.a.track
                            end,
                            get = function()
                                return Scorchio2.db.profile.baroptions.a.show
                            end,
                            set = function(info, v)
                                Scorchio2.db.profile.baroptions.a.show = v
                                Scorchio2:UpdateAnchors()
                            end
                        },
                        bartext = {
                            type = 'input',
                            name = L["Bar Text String"],
                            desc = L["Enter text to appear on Scorch Bars. $s will be replaced by the scorch stack. $m will be replaced by the mob name. Everything else will remain as-is."],
                            width = "full",
                            order = 133,
                            get = function()
                                return Scorchio2.db.profile.baroptions.a.bar
                            end,
                            set = function(info, v)
                                Scorchio2.db.profile.baroptions.a.bar = v
                                Scorchio2:UpdateAnchors()
                            end
                        },                            
                        warningtime = {
                            type = 'range',
                            name = L["Warning Time"],
                            desc = L["Warning Time for Scorch Bars"],
                            order = 134,
                            disabled = function()
                                return not Scorchio2.db.profile.baroptions.a.track
                            end,
                            get = function()
                                return Scorchio2.db.profile.baroptions.a.warningtime
                            end,
                            set = function(info, v)
                                Scorchio2.db.profile.baroptions.a.warningtime = v
                                Scorchio2:UpdateAnchors()                   
                            end,
                            min = 0,
                            max = 30,
                            step = 1,
                        },
                        on = {
                            type = 'toggle',
                            name = L["Toggle Sounds"],
                            desc = L["Enable/Disable sounds for Scorch"],
                            order = 135,
                            disabled = function()
                                return not Scorchio2.db.profile.baroptions.a.track
                            end,
                            get = function()
                                return Scorchio2.db.profile.baroptions.a.soundson
                            end,
                            set = function()
                                Scorchio2.db.profile.baroptions.a.soundson = not Scorchio2.db.profile.baroptions.a.soundson
                            end
                        },
                        warningsound = {
                            type = 'select',
                            name = L["Warning Sound"],
                            desc = L["Change the sound played when warning time is reached"],
                            order = 136,
                            disabled = function()
                                return not Scorchio2.db.profile.baroptions.a.soundson or not Scorchio2.db.profile.baroptions.a.track
                            end,
                            values = SM:List("sound"),
                            get = function() 
                                return GetSMIndex("sound", Scorchio2.db.profile.baroptions.a.warningsound)
                            end,
                            set = function(info, v)
                                Scorchio2.db.profile.baroptions.a.warningsound = SM:List("sound")[v]
                            end,
                        },
                        expiredsound = {
                            type = 'select',
                            name = L["Expired Sound"],
                            desc = L["Change the sound played when the debuff expires"],
                            order = 137,
                            disabled = function()
                                return not Scorchio2.db.profile.baroptions.a.soundson or not Scorchio2.db.profile.baroptions.a.track
                            end,
                            values = SM:List("sound"),
                            get = function() 
                                return GetSMIndex("sound", Scorchio2.db.profile.baroptions.a.expiredsound)
                            end,
                            set = function(info, v)
                                Scorchio2.db.profile.baroptions.a.expiredsound = SM:List("sound")[v]
                            end,
                        },
                        barcolour = {
                            type = 'color',
                            name = L["Bar Colour"],
                            desc = L["Change bar colour"],
                            order = 138,
                            disabled = function()
                                return not Scorchio2.db.profile.baroptions.a.track
                            end,
                            get = function(info)
                                return unpack(Scorchio2.db.profile.baroptions.a.fg)
                            end,
                            set = function(info, r, g, b)
                                Scorchio2.db.profile.baroptions.a.fg = {r, g, b}
                                Scorchio2:UpdateAnchors()
                            end,
                            hasAlpha = false
                        },
                        backgroundcolour = {
                            type = 'color',
                            name = L["Background Colour"],
                            desc = L["Change background colour"],
                            order = 139,
                            disabled = function()
                                return not Scorchio2.db.profile.baroptions.a.track
                            end,
                            get = function(info)
                                return unpack(Scorchio2.db.profile.baroptions.a.bg)
                            end,
                            set = function(info, r, g, b)
                                Scorchio2.db.profile.baroptions.a.bg = {r, g, b}
                                Scorchio2:UpdateAnchors()
                            end,
                            hasAlpha = false
                        }, 
                    },
                },
                b = {
                    type = 'group',
                    name = L["Winter's Chill Bar Options"],
                    desc = L["Winter's Chill Bar Options"],
                    args = {
                        track = {
                            type = 'toggle',
                            name = L["Track Winter's Chill"],
                            desc = L["Track Winter's Chill"],
                            order = 141,
                            get = function()
                                return Scorchio2.db.profile.baroptions.b.track
                            end,
                            set = function(info, v)
                                Scorchio2.db.profile.baroptions.b.track = v
                                Scorchio2:UpdateAnchors()
                            end
                        }, 
                        show = {
                            type = 'toggle',
                            name = L["Show Winter's Chill Bars"],
                            desc = L["Show Winter's Chill Bars"],
                            order = 142,
                            disabled = function()
                                return not Scorchio2.db.profile.baroptions.b.track
                            end,
                            get = function()
                                return Scorchio2.db.profile.baroptions.b.show
                            end,
                            set = function(info, v)
                                Scorchio2.db.profile.baroptions.b.show = v
                                Scorchio2:UpdateAnchors()
                            end
                        },
                        bartext = {
                            type = 'input',
                            name = L["Bar Text String"],
                            desc = L["Enter text to appear on Winter's Chill Bars. $s will be replaced by the scorch stack. $m will be replaced by the mob name. Everything else will remain as-is."],
                            width = "full",
                            order = 143,
                            get = function()
                                return Scorchio2.db.profile.baroptions.b.bar
                            end,
                            set = function(info, v)
                                Scorchio2.db.profile.baroptions.b.bar = v
                                Scorchio2:UpdateAnchors()
                            end
                        },  
                        warningtime = {
                            type = 'range',
                            name = L["Warning Time"],
                            desc = L["Warning Time for Winter's Chill Bars"],
                            order = 144,
                            disabled = function()
                                return not Scorchio2.db.profile.baroptions.b.track
                            end,
                            get = function()
                                return Scorchio2.db.profile.baroptions.b.warningtime
                            end,
                            set = function(info, v)
                                Scorchio2.db.profile.baroptions.b.warningtime = v
                                Scorchio2:UpdateAnchors()                   
                            end,
                            min = 0,
                            max = 30,
                            step = 1,
                        },
                        on = {
                            type = 'toggle',
                            name = L["Toggle Sounds"],
                            desc = L["Enable/Disable sounds for Winter's Chill"],
                            order = 145,
                            disabled = function()
                                return not Scorchio2.db.profile.baroptions.b.track
                            end,
                            get = function()
                                return Scorchio2.db.profile.baroptions.b.soundson
                            end,
                            set = function()
                                Scorchio2.db.profile.baroptions.b.soundson = not Scorchio2.db.profile.baroptions.b.soundson
                            end
                        },
                        warningsound = {
                            type = 'select',
                            name = L["Warning Sound"],
                            desc = L["Change the sound played when warning time is reached"],
                            order = 146,
                            disabled = function()
                                return not Scorchio2.db.profile.baroptions.b.soundson or not Scorchio2.db.profile.baroptions.b.track
                            end,
                            values = SM:List("sound"),
                            get = function() 
                                return GetSMIndex("sound", Scorchio2.db.profile.baroptions.b.warningsound)
                            end,
                            set = function(info, v)
                                Scorchio2.db.profile.baroptions.b.warningsound = SM:List("sound")[v]
                            end,
                        },
                        expiredsound = {
                            type = 'select',
                            name = L["Expired Sound"],
                            desc = L["Change the sound played when the debuff expires"],
                            order = 147,
                            disabled = function()
                                return not Scorchio2.db.profile.baroptions.b.soundson or not Scorchio2.db.profile.baroptions.b.track
                            end,
                            values = SM:List("sound"),
                            get = function() 
                                return GetSMIndex("sound", Scorchio2.db.profile.baroptions.b.expiredsound)
                            end,
                            set = function(info, v)
                                Scorchio2.db.profile.baroptions.b.expiredsound = SM:List("sound")[v]
                            end,
                        },
                        barcolour = {
                            type = 'color',
                            name = L["Bar Colour"],
                            desc = L["Change bar colour"],
                            order = 148,
                            disabled = function()
                                return not Scorchio2.db.profile.baroptions.b.track
                            end,
                            get = function(info)
                                return unpack(Scorchio2.db.profile.baroptions.b.fg)
                            end,
                            set = function(info, r, g, b)
                                Scorchio2.db.profile.baroptions.b.fg = {r, g, b}
                                Scorchio2:UpdateAnchors()
                            end,
                            hasAlpha = false
                        },
                        backgroundcolour = {
                            type = 'color',
                            name = L["Background Colour"],
                            desc = L["Change background colour"],
                            order = 149,
                            disabled = function()
                                return not Scorchio2.db.profile.baroptions.b.track
                            end,
                            get = function(info)
                                return unpack(Scorchio2.db.profile.baroptions.b.bg)
                            end,
                            set = function(info, r, g, b)
                                Scorchio2.db.profile.baroptions.b.bg = {r, g, b}
                                Scorchio2:UpdateAnchors()
                            end,
                            hasAlpha = false
                        }, 
                    },
                },
                c = {
                    type = 'group',
                    name = L["Living Bomb Bar Options"],
                    desc = L["Living Bomb Bar Options"],
                    args = {
                        track = {
                            type = 'toggle',
                            name = L["Track Living Bomb"],
                            desc = L["Track Living Bomb"],
                            order = 151,
                            get = function()
                                return Scorchio2.db.profile.baroptions.c.track
                            end,
                            set = function(info, v)
                                Scorchio2.db.profile.baroptions.c.track = v
                                Scorchio2:UpdateAnchors()
                            end
                        }, 
                        show = {
                            type = 'toggle',
                            name = L["Show Living Bomb Bars"],
                            desc = L["Show Living Bomb Bars"],
                            order = 152,
                            disabled = function()
                                return not Scorchio2.db.profile.baroptions.c.track
                            end,
                            get = function()
                                return Scorchio2.db.profile.baroptions.c.show
                            end,
                            set = function(info, v)
                                Scorchio2.db.profile.baroptions.c.show = v
                                Scorchio2:UpdateAnchors()
                            end
                        },
                        bartext = {
                            type = 'input',
                            name = L["Bar Text String"],
                            desc = L["Enter text to appear on Living Bomb Bars. $s will be replaced by the scorch stack. $m will be replaced by the mob name. Everything else will remain as-is."],
                            width = "full",
                            order = 153,
                            get = function()
                                return Scorchio2.db.profile.baroptions.c.bar
                            end,
                            set = function(info, v)
                                Scorchio2.db.profile.baroptions.c.bar = v
                                Scorchio2:UpdateAnchors()
                            end
                        },  
                        warningtime = {
                            type = 'range',
                            name = L["Warning Time"],
                            desc = L["Warning Time for Living Bomb Bars"],
                            order = 154,
                            disabled = function()
                                return not Scorchio2.db.profile.baroptions.c.track
                            end,
                            get = function()
                                return Scorchio2.db.profile.baroptions.c.warningtime
                            end,
                            set = function(info, v)
                                Scorchio2.db.profile.baroptions.c.warningtime = v
                                Scorchio2:UpdateAnchors()                   
                            end,
                            min = 0,
                            max = 12,
                            step = 1,
                        },
                        on = {
                            type = 'toggle',
                            name = L["Toggle Sounds"],
                            desc = L["Enable/Disable sounds for Living Bomb"],
                            order = 155,
                            disabled = function()
                                return not Scorchio2.db.profile.baroptions.c.track
                            end,
                            get = function()
                                return Scorchio2.db.profile.baroptions.c.soundson
                            end,
                            set = function()
                                Scorchio2.db.profile.baroptions.c.soundson = not Scorchio2.db.profile.baroptions.c.soundson
                            end
                        },
                        warningsound = {
                            type = 'select',
                            name = L["Warning Sound"],
                            desc = L["Change the sound played when warning time is reached"],
                            order = 156,
                            disabled = function()
                                return not Scorchio2.db.profile.baroptions.c.soundson or not Scorchio2.db.profile.baroptions.c.track
                            end,
                            values = SM:List("sound"),
                            get = function() 
                                return GetSMIndex("sound", Scorchio2.db.profile.baroptions.c.warningsound)
                            end,
                            set = function(info, v)
                                Scorchio2.db.profile.baroptions.c.warningsound = SM:List("sound")[v]
                            end,
                        },
                        expiredsound = {
                            type = 'select',
                            name = L["Expired Sound"],
                            desc = L["Change the sound played when the debuff expires"],
                            order = 157,
                            disabled = function()
                                return not Scorchio2.db.profile.baroptions.c.soundson or not Scorchio2.db.profile.baroptions.c.track
                            end,
                            values = SM:List("sound"),
                            get = function() 
                                return GetSMIndex("sound", Scorchio2.db.profile.baroptions.c.expiredsound)
                            end,
                            set = function(info, v)
                                Scorchio2.db.profile.baroptions.c.expiredsound = SM:List("sound")[v]
                            end,
                        },
                        barcolour = {
                            type = 'color',
                            name = L["Bar Colour"],
                            desc = L["Change bar colour"],
                            order = 158,
                            disabled = function()
                                return not Scorchio2.db.profile.baroptions.c.track
                            end,
                            get = function(info)
                                return unpack(Scorchio2.db.profile.baroptions.c.fg)
                            end,
                            set = function(info, r, g, b)
                                Scorchio2.db.profile.baroptions.c.fg = {r, g, b}
                                Scorchio2:UpdateAnchors()
                            end,
                            hasAlpha = false
                        },
                        backgroundcolour = {
                            type = 'color',
                            name = L["Background Colour"],
                            desc = L["Change background colour"],
                            order = 159,
                            disabled = function()
                                return not Scorchio2.db.profile.baroptions.c.track
                            end,
                            get = function(info)
                                return unpack(Scorchio2.db.profile.baroptions.c.bg)
                            end,
                            set = function(info, r, g, b)
                                Scorchio2.db.profile.baroptions.c.bg = {r, g, b}
                                Scorchio2:UpdateAnchors()
                            end,
                            hasAlpha = false
                        }, 
                    },
                },
            },
        },
        output = Scorchio2:GetSinkAce3OptionsDataTable(),
        debug = {
            type = 'group',
            name = "Debug",
            desc = "Debug Options",
            args = {
                duplicates = {
                    type = 'toggle',
                    name = "Duplicates",
                    desc = "Notify of duplicate supression.",
                    get = function()
                        return Scorchio2.db.profile.debug.duplicates
                    end,
                    set = function()
                        Scorchio2.db.profile.debug.duplicates = not Scorchio2.db.profile.debug.duplicates
                    end
                },
                sync = {
                    type = 'toggle',
                    name = "Sync Info",
                    desc = "Display synchronisation debug information",
                    get = function()
                        return Scorchio2.db.profile.debug.sync
                    end,
                    set = function()
                        Scorchio2.db.profile.debug.sync = not Scorchio2.db.profile.debug.sync
                    end                   
                },
                mob = {
                    type = 'toggle',
                    name = "Mob Info",
                    desc = "Display mob debug information",
                    get = function()
                        return Scorchio2.db.profile.debug.mob
                    end,
                    set = function()
                        Scorchio2.db.profile.debug.mob = not Scorchio2.db.profile.debug.mob
                    end
                },
                code = {
                    type = 'toggle',
                    name = "Code Info",
                    desc = "Display code debug information",
                    get = function()
                        return Scorchio2.db.profile.debug.code
                    end,
                    set = function()
                        Scorchio2.db.profile.debug.code = not Scorchio2.db.profile.debug.code
                    end
                },
                timers = {
                    type = 'toggle',
                    name = "Timers Info",
                    desc = "Display timers debug information",
                    get = function()
                        return Scorchio2.db.profile.debug.timers
                    end,
                    set = function()
                        Scorchio2.db.profile.debug.timers = not Scorchio2.db.profile.debug.timers
                    end
                },
                log = {
                    type = 'toggle',
                    name = "Combat Log Info",
                    desc = "Display combat log information",
                    get = function()
                        return Scorchio2.db.profile.debug.log
                    end,
                    set = function()
                        Scorchio2.db.profile.debug.log = not Scorchio2.db.profile.debug.log
                    end
                },                  
            },
        },
    }, -- closes top level group args
} -- closes options

AceConfig:RegisterOptionsTable("Scorchio2", options)

-- Add-on Functions
function Scorchio2:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("Scorchio2DB", defaults, "Default")

	self:RegisterChatCommand("s2", function() LibStub("AceConfigDialog-3.0"):Open("Scorchio2") end )
	local optFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("Scorchio2", "Scorchio2")
	options.args.profile = LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db)
	self:SetSinkStorage(self.db.profile.sinkoptions)
	
	self.db.RegisterCallback(self, "OnProfileCopied", "UpdateAnchors")
	self.db.RegisterCallback(self, "OnProfileReset", "UpdateAnchors")	

    SM.RegisterCallback(self, "LibSharedMedia_Registered", "UpdateAnchors")
    SM.RegisterCallback(self, "LibSharedMedia_SetGlobal", "UpdateAnchors")
    
    self:CreateAnchor("main")
    
    lastTargetGUID = "0"
    mobTable = { }
end

function Scorchio2:OnEnable()
    self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
    self:RegisterEvent("PLAYER_REGEN_ENABLED")
    self:RegisterEvent("PLAYER_TARGET_CHANGED")
end

function Scorchio2:OnDisable()
	self:UnregisterAllEvents()
end


-- Bar and Anchor Functions
function Scorchio2:CreateAnchor(anchor)
	self[anchor] = self:NewBarGroup(anchor, nil, self.db.profile.baroptions.width, self.db.profile.baroptions.height)
	self[anchor]:ClearAllPoints()
	
	self[anchor].id = anchor
	self[anchor].RegisterCallback(self, "AnchorClicked")
	self[anchor]:HideAnchor()
	self[anchor]:SetSortFunction(Scorchio2.SortBars)
	self:UpdateAnchor(anchor)
end

function Scorchio2:RunTest()
	local icon = select(3, GetSpellInfo(27074)) or nil
	local start = GetTime()
    Scorchio2:UpdateTable("Test1", "Test Mob 1", 1, GetTime(), 5)
    Scorchio2:UpdateTable("Test2", "Test Mob 2", 3, GetTime(), 1)
    Scorchio2:UpdateTable("Test3", "Test Mob 3", 2, GetTime(), 3)
end

function Scorchio2:AnchorClicked(callback, group, button)
	self:SavePosition(group.id)
	if button == "RightButton" then
		group:HideAnchor()
		group:Lock()
	end
end
function Scorchio2:ToggleAnchors()
	if self.main.button:IsVisible() then
		self.main:HideAnchor()
		self.main:Lock()
	else
		self.main:ShowAnchor()
		self.main:Unlock()
	end
end

function Scorchio2:SavePosition(anchor)
	local p,_,pr,px,py = self[anchor]:GetPoint()
	self.db.profile.baroptions.position[anchor].p = p
	self.db.profile.baroptions.position[anchor].pr = pr
	self.db.profile.baroptions.position[anchor].px = px
	self.db.profile.baroptions.position[anchor].py = py
end

function Scorchio2:UpdateAnchor(anchor)
	self[anchor]:SetTexture(SM:Fetch("statusbar", self.db.profile.baroptions.texture))
	self[anchor]:SetFont(SM:Fetch("font", self.db.profile.baroptions.font), self.db.profile.baroptions.fontsize)
	self[anchor]:SetWidth(self.db.profile.baroptions.width)
	self[anchor]:SetHeight(self.db.profile.baroptions.height)
	self[anchor]:ReverseGrowth(self.db.profile.baroptions.growth)
    if self.db.profile.baroptions.maxbars == 0 then
        self[anchor]:SetMaxBars(nil)
    else
        self[anchor]:SetMaxBars(self.db.profile.baroptions.maxbars)
    end
	self[anchor]:SetFlashPeriod(0)

	if self.db.profile.baroptions.flash then
		self[anchor]:SetFlashPeriod(self.db.profile.baroptions.flashtime)
	end

	if not self.db.profile.baroptions.position[anchor] then self.db.profile.baroptions.position[anchor] = {} end
	if self.db.profile.baroptions.position[anchor].p
       and self.db.profile.baroptions.position[anchor].pr
       and self.db.profile.baroptions.position[anchor].px
       and self.db.profile.baroptions.position[anchor].py then
		self[anchor]:ClearAllPoints()
		self[anchor]:SetPoint(self.db.profile.baroptions.position[anchor].p, UIParent, self.db.profile.baroptions.position[anchor].pr, self.db.profile.baroptions.position[anchor].px, self.db.profile.baroptions.position[anchor].py)
		else
			self[anchor]:SetPoint("CENTER", UIParent, "CENTER", 0, 0)
		end
end

function Scorchio2:UpdateAnchors()
	self:UpdateAnchor("main")
end

function Scorchio2.SortBars(a, b)
	if a.isTimer ~= b.isTimer then
		return a.isTimer
	end
	
	if a.maxValue == b.maxValue then
		if a.value == b.value then
			return a.name > b.name
		else
			return a.value > b.value
		end
	else
		return a.maxValue > b.maxValue
	end
end


-- Event Handling Functions
function Scorchio2:COMBAT_LOG_EVENT_UNFILTERED(_, _, event, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, ...)
    local _, _, prefix, suffix = string.find(event, "(.-)_(.+)")

    if not InCombatLockdown("player") then return end

    local timeStamp = GetTime()

   Debug("log", "LOG: Time: " .. tostring(timeStamp) .. ", Event: " .. tostring(event) .. ", dstName: " .. tostring(dstName) .. ", dstFlags: " .. tostring(dstFlags))
    
	if (prefix == "SPELL") then
		local spellID, spellName, spellSchool, spellAuraType, spellCount, fvApplied
		spellCount = -1
		spellID = arg9
--~ 		if (suffix == "DAMAGE") then
--~             for k, v in pairs(SPELLIDS) do
--~                 if (v[tonumber(spellID)]) then
--~                     if Scorchio2.db.profile.baroptions[T[k]].track then
--~                         self:UpdateTable(dstGUID, dstName, k, timeStamp, nil)
--~                     end
--~                 end
--~             end
--~         else
--~         if mobTable[dstGUID] ~= nil then
            if (suffix == "AURA_APPLIED") then
                spellCount = 1
                self:CheckVuln(arg9, spellCount, timeStamp, dstGUID, dstName, srcFlags)
            elseif (suffix == "AURA_APPLIED_DOSE") then
                self:CheckVuln(arg9, arg13, timeStamp, dstGUID, dstName, srcFlags)
            elseif (suffix == "AURA_REFRESH") then
                self:CheckVuln(arg9, nil, timeStamp, dstGUID, dstName, srcFlags)
            elseif (suffix == "AURA_REMOVED") then		
                spellCount = -1
                self:CheckVuln(arg9, spellCount, timeStamp, dstGUID, dstName, srcFlags)
            elseif (suffix == "AURA_REMOVED_DOSE") then						
                self:CheckVuln(arg9, arg13, timeStamp, dstGUID, dstName, srcFlags)			
            end
--~         end
	elseif (prefix == "UNIT") then
		if dstName ~= nil and mobTable[dstGUID] ~= nil then
            self:MobDies(dstGUID, dstName)
		end
	end
end

function Scorchio2:PLAYER_REGEN_ENABLED()
    for k, v in pairs(mobTable) do
        for kk = 1,table.getn(T) do
            if mobTable[k][kk] then
                if mobTable[k][kk][BAR] then
                    mobTable[k][kk][BAR]:Fade(0.5)
                    Scorchio2:ReleaseBar(mobTable[k][kk][BAR])
                end
            end
        end
    end
    Scorchio2:CancelAllTimers()  
    mobTable = { }
    lastTargetGUID = "0"
end

function Scorchio2:PLAYER_TARGET_CHANGED()
    if mobTable[lastTargetGUID] then
        for k = 1,table.getn(T) do
            if mobTable[lastTargetGUID][k] then
                if mobTable[lastTargetGUID][k][BAR] then
                    mobTable[lastTargetGUID][k][BAR].texture:SetVertexColor(Colours(Scorchio2.db.profile.baroptions[T[k]].fg, 0.1))
                    mobTable[lastTargetGUID][k][BAR].bgtexture:SetVertexColor(Colours(Scorchio2.db.profile.baroptions[T[k]].bg, 0.1))    
                end
            end
        end
    end
    if UnitExists("target") then
		lastTargetGUID = UnitGUID("target")
        for k = 1,table.getn(T) do
            if mobTable[lastTargetGUID] then
                if mobTable[lastTargetGUID][k] then
                    if mobTable[lastTargetGUID][k][BAR] then
                        mobTable[lastTargetGUID][k][BAR].texture:SetVertexColor(Colours(Scorchio2.db.profile.baroptions[T[k]].fg, 1.0))
                        mobTable[lastTargetGUID][k][BAR].bgtexture:SetVertexColor(Colours(Scorchio2.db.profile.baroptions[T[k]].bg, 1.0))    
                    end
                end
            end
        end
	else
		lastTargetGUID = "0"
	end
end

-- mobTable Handling Functions
function Scorchio2:MobDies(dstGUID, dstName)
    Debug("mob", tostring(dstName) .. " dies.")
    for k = 1,table.getn(T) do
        if mobTable[dstGUID][k] then
            if mobTable[dstGUID][k][BAR] then
                mobTable[dstGUID][k][BAR]:Fade(0.5)
                Scorchio2:ReleaseBar(mobTable[dstGUID][k][BAR])
            end
            if mobTable[dstGUID][k][TIMER] then
                Scorchio2:CancelTimer(mobTable[dstGUID][k][TIMER], true) 
            end
        end
    end
    mobTable[dstGUID] = nil
--~     raidIconTable[self:GetGUID(fullMobName)] = nil
--~     self:UnregisterEnniBarWithGroup(fullMobName,SB)
--~     self:UnregisterEnniBar(fullMobName)
end
function Scorchio2:CheckVuln(spellID, spellCount, timeStamp, dstGUID, dstName, srcFlags)
    for k, v in pairs(VULNERABILITIES) do
        if v[spellID] then
            if bit.band(srcFlags, VULNDATA[k].valid) ~= 0 and Scorchio2.db.profile.baroptions[T[k]].track then
                self:UpdateTable(dstGUID, dstName, k, timeStamp, spellCount)
            end
        end
    end 
end

function Scorchio2:LostVuln(dstGUID, vulnKey)
--~     Debug("code", "Time Diff: " .. tostring(GetTime() - mobTable[dstGUID][vulnKey][REFRESHTIME])) 
    if GetTime() - mobTable[dstGUID][vulnKey][REFRESHTIME] > ( VULNDATA[vulnKey].duration - 0.25 ) then
        if Scorchio2.db.profile.baroptions[T[vulnKey]].soundson then
            PlaySoundFile(SM:Fetch("sound", Scorchio2.db.profile.baroptions[T[vulnKey]].expiredsound))
        end
        self:Pour(string.format(VULNDATA[vulnKey].message, tostring(mobTable[dstGUID][vulnKey][MOBNAME])), 1, 0, 0)
        Scorchio2:CancelTimer(mobTable[dstGUID][vulnKey][TIMER], true)  
        mobTable[dstGUID][vulnKey] = nil
    --~     if self:IsEnniBarRegistered(fullMobName) then
    --~         self:UnregisterEnniBarWithGroup(fullMobName,SB)
    --~         self:UnregisterEnniBar(fullMobName)
    --~     end 
    end
end

function Scorchio2:TimerComplete(info)
    local dstGUID, vulnKey = info[1], info[2]
	local timerTime = GetTime()
--~     Debug("code","dstGUID=" .. tostring(dstGUID))
--~     Debug("code","mobTable[dstGUID][vulnKey][REFRESHTIME]=" .. tostring(mobTable[dstGUID][vulnKey][REFRESHTIME]))
--~     Debug("code","timerTime=" .. tostring(timerTime))
--~     Debug("code","timerTime - mobTable[dstGUID][vulnKey][REFRESHTIME]=" .. tostring(timerTime - mobTable[dstGUID][vulnKey][REFRESHTIME]))
    if mobTable[dstGUID][vulnKey] then
		Debug("timers", "TIMER: " ..tostring(dstGUID) .. ", " .. tostring(mobTable[dstGUID][vulnKey][REFRESHTIME]) .. ", " .. tostring(timerTime) .. ", " .. tostring(timerTime - mobTable[dstGUID][vulnKey][REFRESHTIME]))
		if math.floor((timerTime - mobTable[dstGUID][vulnKey][REFRESHTIME]) + 0.5) >= (VULNDATA[vulnKey].duration - Scorchio2.db.profile.baroptions[T[vulnKey]].warningtime) then
			if Scorchio2.db.profile.baroptions[T[vulnKey]].soundson then
				PlaySoundFile(SM:Fetch("sound", Scorchio2.db.profile.baroptions[T[vulnKey]].warningsound))
			end
			self:Pour(string.format(VULNDATA[vulnKey].warning, tostring(mobTable[dstGUID][vulnKey][MOBNAME])), 1, 0.647, 0)
		end
	end
end

function Scorchio2:UpdateTable(dstGUID, dstName, vulnKey, timeStamp, spellCount)
--~     Debug("code","dstGUID=" .. tostring(dstGUID) .. ", GUID of Target=" .. UnitGUID("target") .. ", lastTargetGUID=" .. tostring(lastTargetGUID))
--~     Debug("code","dstName=" .. tostring(dstName))
--~     Debug("code","vulnKey=" .. tostring(vulnwKey))
--~     Debug("code","timeStamp=" .. tostring(timeStamp))
--~     Debug("code","spellCount=" .. tostring(spellCount))
	if mobTable[dstGUID] == nil then
        Debug("mob", "New mob: " .. tostring(dstGUID) .. "_" .. tostring(dstName) .. " VulnKey: " .. tostring(vulnKey))
        mobTable[dstGUID] = { }
	end
    if mobTable[dstGUID][vulnKey] == nil then
        Debug("mob", "New vulnerability: " .. tostring(dstGUID) .. "_" .. tostring(dstName) .. " VulnKey: " .. tostring(vulnKey))
        table.insert(mobTable[dstGUID], vulnKey, {dstName, timeStamp, 0, 0, nil })
    end
    Debug("mob","MOB: dstGUID=" .. tostring(dstGUID) .. ", vulnKey=" .. tostring(vulnKey) .. ", spellCount=" .. tostring(spellCount))
    if spellCount == nil then spellCount = mobTable[dstGUID][vulnKey][STACK] + 1 end
    if spellCount > VULNDATA[vulnKey].maxstacks then spellCount = VULNDATA[vulnKey].maxstacks end
    if spellCount <= 0 then
        self:LostVuln(dstGUID, vulnKey)
    else
        local warningTime = Scorchio2.db.profile.baroptions[T[vulnKey]].warningtime
        local icon = VULNDATA[vulnKey].icon
        local duration = VULNDATA[vulnKey].duration
        local alpha = 1
        mobTable[dstGUID][vulnKey][STACK] = spellCount
        mobTable[dstGUID][vulnKey][REFRESHTIME] = timeStamp
        if warningTime > 0 then
            if mobTable[dstGUID][vulnKey][TIMER] ~= nil then
                Scorchio2:CancelTimer(mobTable[dstGUID][vulnKey][TIMER], true)
            end
            mobTable[dstGUID][vulnKey][TIMER] = Scorchio2:ScheduleTimer("TimerComplete", (duration - warningTime), { dstGUID , vulnKey })
        end
        if tostring(dstGUID) == UnitGUID("target") then
            lastTargetGUID = dstGUID
        else
            alpha = 0.1
        end
        if Scorchio2.db.profile.baroptions[T[vulnKey]].show then
            if VULNDATA[vulnKey].unique then
               for k, v in pairs(mobTable) do
                    if mobTable[k][vulnKey] then
                        if mobTable[k][vulnKey][BAR] then
                            mobTable[k][vulnKey][BAR]:Fade(0.5)
                            Scorchio2:ReleaseBar(mobTable[k][vulnKey][BAR])
                        end
                        if mobTable[k][vulnKey][TIMER] then
                            Scorchio2:CancelTimer(mobTable[k][vulnKey][TIMER], true)
                        end
                    end
                end
            end
            local barText = string.gsub(Scorchio2.db.profile.baroptions[T[vulnKey]].bar, '$s', tostring(spellCount))
            barText = string.gsub(barText, '$m', tostring(dstName))
            mobTable[dstGUID][vulnKey][BAR] = Scorchio2.main:NewTimerBar(tostring(dstGUID .. "_" .. vulnKey), barText, duration - (GetTime() - timeStamp), duration, icon)
            mobTable[dstGUID][vulnKey][BAR].texture:SetVertexColor(Colours(Scorchio2.db.profile.baroptions[T[vulnKey]].fg, alpha))
            mobTable[dstGUID][vulnKey][BAR].bgtexture:SetVertexColor(Colours(Scorchio2.db.profile.baroptions[T[vulnKey]].bg, alpha))
        end
    end
end